#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import os
import sys
import numpy
import _pickle as pickle

import ase
from ase              import Atom, Atoms
from ase.spacegroup   import crystal
from ase.data         import chemical_symbols
from ase.neighborlist import NeighborList
from collections import Counter

from scipy.cluster.hierarchy import fclusterdata as cluster

import warnings

from autografs.utils.pointgroup import PointGroup
from autografs.utils           import __data__

import logging
logger = logging.getLogger(__name__)


warnings.filterwarnings("error")

class Topology(object):
    """Contener class for the topology information"""

    def __init__(self,
                 name  : str,
                 atoms : ase.Atoms) -> None:
        """Constructor for a topology, from an ASE Atoms."""
        logger.debug("Creating Topology {0}".format(name))
        self.name  = name
        self.atoms = atoms 
        # initialize empty fragments
        # shapes and symmops will be used to find
        # corresponding SBUs.
        self.fragments = {}
        self.shapes    = {}
        self.symmops   = {}
        # fill it in
        self._analyze()
        return None

    def get_atoms(self) -> ase.Atoms:
        """Return a copy of the topology as ASE Atoms."""
        logger.debug("Topology {0}: returning atoms.".format(self.name))
        return self.atoms.copy()

    def get_fragments(self) -> ase.Atoms:
        """Return a concatenated version of the fragments."""
        logger.debug("Topology {0}: returning fragments.".format(self.name))
        frags = ase.Atoms(cell=self.atoms.get_cell(),
                          pbc=self.atoms.get_pbc())
        for idx,frag in self.fragments.items():
            tags = numpy.ones(len(frag))*idx
            frag.set_tags(tags)
            frags+=frag
        return frags

    def get_unique_shapes(self) -> set:
        """Return all unique shapes in the topology."""
        logger.debug("Topology {0}: listing unique fragment shapes.".format(self.name))
        return set(self.shapes.values())

    def get_unique_operations(self) -> dict:
        """Return all unique symmetry operations in the topology."""
        logger.debug("Topology {0}: listing symmetry operations.".format(self.name))
        ops = {}
        for idx,shape in self.shapes.items():
            if shape in ops.keys():
                continue
            symmops = self.symmops[idx]
            these_ops = []
            for s,o in symmops.items():
                if o is None:
                    continue
                elif len(o)>0 and s not in ["I","-I"]:
                    sym = ["{0}{1}".format(s,this_o[0]) for this_o in o]
                elif len(o)>0 and o is not None:
                    sym = [s,]
                else:
                    sym = []
                these_ops += sym
            ops[shape] = {o:these_ops.count(o) for o in set(these_ops)}
        return ops

    def has_compatible_slot(self,
                            sbu : object) -> (bool,tuple):
        """Return (True,shape) for a slot compatible with the SBU"""
        compatible = False
        slot      = None
        if sbu.shape in self.get_unique_shapes():
            compatible = True
            slot       = sbu.shape
        else:
            symmops = self.get_unique_operations()
            for shape,ops in symmops.items():
                if shape[1]!=sbu.shape[1]:
                    compatible = False
                    slot      = None
                    break
                elif sbu.is_compatible(ops):
                    compatible = True
                    slot       = shape
                    break
        return compatible,slot

    def _get_cutoffs(self,
                     Xis : numpy.ndarray,
                     Ais : numpy.ndarray) -> numpy.ndarray:
        """Return the cutoffs leading to the desired connectivity"""
        # initialize cutoffs to small non-zero skin partameter
        # TODO check on coordination
        logger.debug("Generating cutoffs")
        skin    = 5e-3
        cutoffs = numpy.zeros(len(self.atoms))+skin
        # we iterate over non-dummies
        for other_index in Ais:
            # cutoff starts impossibly big
            cutoff   = 10000.0
            # we get the distances to all dummies and cluster accordingly
            dists    = self.atoms.get_distances(other_index,Xis,mic=True)
            mindist  = numpy.amin(dists)
            # let's not analyse every single distance...
            dists    = dists[dists<2.0*mindist].reshape(-1,1)
            L   = 0
            eps = mindist*0.5
            coord = self.atoms[other_index].number
            # we nee to coerce the cutoffs to have 
            # the good amount of dummies. 
            # in theory, should loop only once
            while L!=coord:
                clusters = cluster(dists, eps, criterion='distance')
                for cluster_index in set(clusters):
                    # check this cluster distances
                    indices   = numpy.where(clusters==cluster_index)[0]
                    L = len(indices)
                    if L>coord:
                        eps*=0.75
                    elif L<coord:
                        eps*=1.5
                    cutoff_tmp = dists[indices].mean() 
                    if cutoff_tmp<cutoff :
                        # if better score, replace the cutoff
                        cutoff = cutoff_tmp
            cutoffs[other_index] = cutoff
        return cutoffs

    def _analyze(self) -> None:
        """Analyze the topology to cut the fragments out."""
        # separate the dummies from the rest
        logger.debug("Analyzing fragments of topology {0}.".format(self.name))
        numbers = numpy.asarray(self.atoms.get_atomic_numbers())
        Xis  = numpy.where(numbers==0)[0]
        Ais  = numpy.where(numbers >0)[0]
        # setup the tags
        tags = numpy.zeros(len(self.atoms))
        tags[Xis] = Xis + 1
        self.atoms.set_tags(tags)
        tags = self.atoms.get_tags()
        # analyze
        # first build the neighborlist
        cutoffs      = self._get_cutoffs(Xis=Xis,Ais=Ais) 
        neighborlist = NeighborList(cutoffs=cutoffs,
                                    bothways=True,
                                    self_interaction=False,
                                    skin=0.0)
        neighborlist.build(self.atoms)
        # iterate over non-dummies to find dummy neighbors
        for ai in Ais:
            # get indices and offsets of dummies only!
            ni,no = neighborlist.get_neighbors(ai)
            ni,no = zip(*[(idx,off) 
                          for idx,off in list(zip(ni,no)) if idx in Xis])
            ni = numpy.asarray(ni)
            no = numpy.asarray(no)
            # get absolute positions, no offsets
            positions = self.atoms.positions[ni] + no.dot(self.atoms.cell)
            # create the Atoms object
            fragment = Atoms("X"*len(ni),positions,tags=tags[ni]) 
            # calculate the point group properties
            pg       = PointGroup(fragment.copy(),cutoffs.mean())
            # save that info
            self.fragments[ai] = fragment
            self.shapes[ai]    = (pg.schoenflies,self.atoms[ai].number)
            self.symmops[ai]   = pg.symmops
        return None



def download_topologies() -> None:
    """Downloads the topology file from the RCSR website"""
    import requests
    import shutil
    url  ="http://rcsr.anu.edu.au/downloads/RCSRnets.cgd"
    root = os.path.join(__data__,"topologies")
    path = os.path.join(root,"nets.cgd")
    resp = requests.get(url, stream=True)
    if resp.status_code == 200:
        logger.info("Successfully downloaded the nets from RCSR.")
        resp.raw.decode_content = True
        with open(path,"wb") as outpt:
            shutil.copyfileobj(resp.raw, outpt)        
    return

def read_topologies_database(update_db     : bool = False,
                             update_source : bool = False) -> dict:
    """Return a dictionary of topologies as ASE Atoms."""
    from autografs.utils.io import read_cgd
    root     = os.path.join(__data__,"topologies")
    db_file  = os.path.join(root,"topologies.pkl")
    cgd_file = os.path.join(root,"nets.cgd")
    if ((not os.path.isfile(cgd_file)) or (update_source)):
        logger.info("Downloading the topologies from RCSR.")
        download_topologies()
    else:
        logger.info("Using saved nets from RCSR")
    if ((not os.path.isfile(db_file)) or (update_db)):
        logger.info("Reloading the topologies from scratch")
        topologies     = read_cgd()
        topologies_len = len(topologies)
        logger.info("{0:<5} topologies saved".format(topologies_len))
        with open(db_file,"wb") as pkl:
            pickle.dump(obj=topologies,file=pkl)
        return topologies
    else:
        logger.info("Using saved topologies")
        with open(db_file,"rb") as pkl:
            topologies     = pickle.load(file=pkl)
            topologies_len = len(topologies)
            logger.info("{0:<5} topologies loaded".format(topologies_len))
            return topologies



if __name__ == "__main__":

    topologies = read_topologies_database(update_db=True,update_source=True)




#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Heavily adapted from the pymatgen library

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat","Shyue Ping Ong"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import numpy as np
from   scipy import cluster
import itertools
from ase.visualize import view
from ase import Atom, Atoms

from autografs.utils.operations import rotation, reflection, inertia


class PointGroup(object):
    """A class to analyze the point group of a molecule. The general outline of
    the algorithm is as follows:

    1. Center the molecule around its center of mass.
    2. Compute the inertia tensor and the eigenvalues and eigenvectors.
    3. Handle the symmetry detection based on eigenvalues.

        a. Linear molecules have one zero eigenvalue. Possible symmetry
           operations are C*v or D*v
        b. Asymetric top molecules have all different eigenvalues. The
           maximum rotational symmetry in such molecules is 2
        c. Symmetric top molecules have 1 unique eigenvalue, which gives a
           unique rotation axis.  All axial point groups are possible
           except the cubic groups (T & O) and I.
        d. Spherical top molecules have all three eigenvalues equal. They
           have the rare T, O or I point groups.

    .. attribute:: schoenflies

        Schoenflies symbol of the detected point group.
    """

    def __init__(self, mol, tol, out=None):
        """
        The default settings are usually sufficient.

        Args:
            mol (Molecule): Molecule to determine point group for.
        """

        self.tol  = tol
        self.etol = self.tol /3.0
        self.mol  = mol
        # center molecule
        self.mol.positions -= self.mol.positions.mean(axis=0)
                # normalize
        norm_factor = np.amax(np.linalg.norm(self.mol.positions,axis=1,keepdims=True))
        if norm_factor<1e-6:
            norm_factor = 1.0
        self.mol.positions /= norm_factor

        self.symmops      = {"C"    :[],
                             "S"    :[],
                             "sigma":[]}
        # identity
        self.symmops["I"] = np.eye(3)
        #inversion
        inversion = - np.eye(3)
        if self.is_valid_op(inversion):
            self.symmops["-I"] = inversion
            # print("Inversion center present")
        else:
            self.symmops["-I"] = None

        self.nrot  = 0
        self.nref  = 0

        self.out = out

        self.analyze()
        
        if self.schoenflies in ["C1v", "C1h"]:
            self.schoenflies = "Cs"

    def analyze(self):
        """TODO"""
        if len(self.mol) == 1:
            self.schoenflies = "Kh"
        else:
            # Get the inertia tensor
            xyz = self.mol.get_positions()
            W   = self.mol.get_masses()
            I = inertia(xyz=xyz,W=W)
            I /= np.sum(np.linalg.norm(xyz,axis=1)*W)
            # calculate the eigenvalues
            eigvals, eigvecs = np.linalg.eigh(I)
            self.eigvecs = eigvecs.T
            self.eigvals = eigvals
            e0,e1,e2 = self.eigvals
            eig_zero = np.abs(np.product(self.eigvals)) < self.etol**3
            eig_same = (abs(e0-e1)<self.etol)&(abs(e0-e2)<self.etol)
            eig_diff = (abs(e0-e1)>self.etol)&(abs(e0-e2)>self.etol)&(abs(e1-e2)>self.etol)
            # print("Inertia tensor eigenvalues = {0:3.2f} {1:3.2f} {2:3.2f}".format(*self.eigvals),file=self.out)
            # determine which process to go through
            if eig_zero:
                # print("Linear molecule detected",file=self.out)
                self.analyze_linear()
            elif eig_same:
                # print("Spherical top molecule detected",file=self.out)
                self.analyze_spherical_top()
            elif eig_diff:
                # print("Asymmetric top molecule detected",file=self.out)
                self.analyze_asymmetric_top()
            else:
                # print("Symmetric top molecule detected",file=self.out)
                self.analyze_symmetric_top()


    def analyze_linear(self):
        """TODO"""
        inversion = - np.eye(3)
        if self.symmops["-I"] is not None:
            self.schoenflies = "D*h"
        else:
            self.schoenflies = "C*v"

    def analyze_asymmetric_top(self):
        """Handles assymetric top molecules, which cannot contain rotational
        symmetry larger than 2.
        """

        for axis in self.eigvecs:
            op = rotation(axis=axis,order=2)
            if self.is_valid_op(op):
                self.symmops["C"] += [(2,axis,op),]
                self.nrot += 1

        if   self.nrot == 0:
            # print("No rotation symmetries detected.",file=self.out)
            self.analyze_nonrotational_groups()
        elif self.nrot == 3:
            # print("Dihedral group detected.",file=self.out)
            self.analyze_dihedral_groups()
        else:
            # print("Cyclic group detected.",file=self.out)
            self.analyze_cyclic_groups()

    def analyze_symmetric_top(self):
        """
        Handles symetric top molecules which has one unique eigenvalue whose
        corresponding principal axis is a unique rotational axis.  More complex
        handling required to look for R2 axes perpendicular to this unique
        axis.
        """

        # identify unique axis
        if   abs(self.eigvals[0] - self.eigvals[1]) < self.etol:
            unique_axis = self.eigvecs[2]
        elif abs(self.eigvals[1] - self.eigvals[2]) < self.etol:
            unique_axis = self.eigvecs[0]
        else:
            unique_axis = self.eigvecs[1]
        order = self.detect_rotational_symmetry(unique_axis)
        # print("Rotation symmetries = {0}".format(order),file=self.out)
        if self.nrot > 0:
            self.has_perpendicular_C2(unique_axis)
        if self.nrot > 1:
            self.analyze_dihedral_groups()
        elif self.nrot == 1:
            self.analyze_cyclic_groups()
        else:
            self.analyze_nonrotational_groups()

    def analyze_nonrotational_groups(self):
        """
        Handles molecules with no rotational symmetry. Only possible point
        groups are C1, Cs and Ci.
        """
        self.schoenflies = "C1"
        if self.symmops["-I"] is not None:
            self.schoenflies = "Ci"
        else:
            for v in self.eigvecs:
                mirror_type = self.find_reflection_plane(v)
                if mirror_type is not None:
                    self.schoenflies = "Cs"
                    break

    def analyze_cyclic_groups(self):
        """
        Handles cyclic group molecules.
        """
        order,axis,op = max(self.symmops["C"], key=lambda v: v[0])
        self.schoenflies = "C{0}".format(order)
        mirror_type = self.find_reflection_plane(axis)
        if mirror_type == "h":
            self.schoenflies += "h"
        elif mirror_type == "v":
            self.schoenflies += "v"
        elif mirror_type is None:
            rotoref = reflection(axis).dot(rotation(axis=axis,order=2*order))
            if self.is_valid_op(rotoref):
                self.schoenflies = "S{0}".format(2*order)
                self.symmops["S"] +=  [(order,axis,rotoref),]

    def analyze_dihedral_groups(self):
        """
        Handles dihedral group molecules, i.e those with intersecting R2 axes
        and a main axis.
        """
        order,axis,op = max(self.symmops["C"], key=lambda v: v[0])
        self.schoenflies = "D{0}".format(order)
        mirror_type = self.find_reflection_plane(axis)
        if mirror_type == "h":
            self.schoenflies += "h"
        elif not mirror_type == "":
            self.schoenflies += "d"

    def find_reflection_plane(self, axis):
        """
        Looks for mirror symmetry of specified type about axis.  Possible
        types are "h" or "vd".  Horizontal (h) mirrors are perpendicular to
        the axis while vertical (v) or diagonal (d) mirrors are parallel.  v
        mirrors has atoms lying on the mirror plane while d mirrors do
        not.
        """
        symmop = None

        # First test whether the axis itself is the normal to a mirror plane.
        op = reflection(axis)
        if self.is_valid_op(op):
            symmop = ("h",axis,op)
        else:
            # Iterate through all pairs of atoms to find mirror
            for s1, s2 in itertools.combinations(self.mol, 2):
                if s1.symbol == s2.symbol:
                    normal = s1.position - s2.position
                    if normal.dot(axis) < self.tol:
                        op = reflection(normal)
                        if self.is_valid_op(op):
                            if self.nrot > 1:
                                symmop = ("d",normal,op)
                                for prev_order,prev_axis,prev_op in self.symmops["C"]:
                                    if not np.linalg.norm(prev_axis - axis) < self.tol:
                                        if np.dot(prev_axis, normal) < self.tol:
                                            symmop = ("v",normal,op)
                                            break
                            else:
                                symmop = ("v",normal,op)
                            break
        if symmop is not None:
            self.symmops["sigma"] += [symmop,]
            return symmop[0]
        else:
            return symmop

    def find_possible_equivalent_positions(self,axis=None):
        """
        Returns the smallest list of atoms with the same species and
        distance from origin AND does not lie on the specified axis.  This
        maximal set limits the possible rotational symmetry operations,
        since atoms lying on a test axis is irrelevant in testing rotational
        symmetryOperations.
        """

        from scipy.cluster.hierarchy import fclusterdata as cluster

        def not_on_axis(index):
            v = np.cross(self.mol.positions[index], axis)
            return np.linalg.norm(v) > self.tol

        valid_sets = []
        numbers  = self.mol.get_atomic_numbers()
        dists    = np.linalg.norm(self.mol.get_positions(),axis=1,keepdims=True)
        clusters = cluster(dists, self.tol, criterion='distance')
        for cval in set(clusters):
            indices = np.where(clusters==cval)[0]
            #most common specie of set only
            counts = np.bincount(numbers[indices])
            indices = indices[np.where(numbers[indices]==np.argmax(counts))[0]]
            if axis is not None:
                indices = list(filter(not_on_axis, indices))
            if len(indices)>1:
                # no symmetry for a set of 1...
                valid_sets.append(self.mol.positions[indices])

        return min(valid_sets, key=lambda s: len(s))

    def detect_rotational_symmetry(self, axis):
        """
        Determines the rotational symmetry about supplied axis.  Used only for
        symmetric top molecules which has possible rotational symmetry
        operations > 2.
        """
        min_set = self.find_possible_equivalent_positions(axis=axis)
        max_sym = len(min_set)
        for order in range(max_sym, 0, -1):
            if max_sym % order != 0:
                continue
            op = rotation(axis=axis,order=order)
            if self.is_valid_op(op,verbose=True):
                # print("Found axis with order {0}".format(order))
                self.symmops["C"] += [(order,axis,op),]
                self.nrot += 1
                return order
        return 1

    def has_perpendicular_C2(self, axis):
        """
        Checks for R2 axes perpendicular to unique axis.  For handling
        symmetric top molecules.
        """
        min_set = self.find_possible_equivalent_positions(axis=axis)
        found=False
        for s1, s2 in itertools.combinations(min_set, 2):
            test_axis = np.cross(s1 - s2, axis)
            if np.linalg.norm(test_axis) > self.tol:
                op = rotation(axis=test_axis,order=2)
                if self.is_valid_op(op):
                    self.symmops["C"] += [(2,test_axis,op),]
                    self.nrot += 1
                    found= True
        return found
        
    def analyze_spherical_top(self):
        """
        Handles Sperhical Top Molecules, which belongs to the T, O or I point
        groups.
        """
        self.find_spherical_axes()
        if self.nrot == 0:
            # print("Accidental spherical top!",file=self.out)
            self.analyze_symmetric_top()
        # get the main axis and order
        order,axis,op = max(self.symmops["C"], key=lambda v: v[0])
        if order < 3:
            # print("Accidental speherical top!",file=self.out)
            self.analyze_symmetric_top()
        elif order == 3:
            mirror_type = self.find_reflection_plane(axis)
            if mirror_type != "":
                if self.symmops["-I"] is not None:
                    self.schoenflies = "Th"
                else:
                    self.schoenflies = "Td"
            else:
                self.schoenflies = "T"
        elif order == 4:
            if self.symmops["-I"] is not None:
                self.schoenflies = "Oh"
            else:
                self.schoenflies = "O"
        elif order == 5:
            if self.symmops["-I"] is not None:
                self.schoenflies = "Ih"
            else:
                self.schoenflies = "I"

    def find_spherical_axes(self):
        """
        Looks for R5, R4, R3 and R2 axes in spherical top molecules.  Point
        group T molecules have only one unique 3-fold and one unique 2-fold
        axis. O molecules have one unique 4, 3 and 2-fold axes. I molecules
        have a unique 5-fold axis.
        """

        rot_present = {2:False,3:False,4:False,5:False}
        test_set = self.find_possible_equivalent_positions()
        for c1, c2, c3 in itertools.combinations(test_set, 3):
            for cc1, cc2 in itertools.combinations([c1, c2, c3], 2):
                if not rot_present[2]:
                    test_axis = cc1 + cc2
                    if np.linalg.norm(test_axis) > self.tol:
                        op = rotation(axis=test_axis,order=2)
                        if self.is_valid_op(op):
                            # print("Found axis with order {0}".format(2))
                            rot_present[2] = True
                            self.symmops["C"] += [(2,test_axis,op),]
                            self.nrot += 1

            test_axis = np.cross(c2 - c1, c3 - c1)
            if np.linalg.norm(test_axis) > self.tol:
                for order in (3, 4, 5):
                    if not rot_present[order]:
                        op = rotation(axis=test_axis,order=order)
                        if self.is_valid_op(op):
                            # print("Found axis with order {0}".format(order))
                            rot_present[order] = True
                            self.symmops["C"] += [(order,test_axis,op),]
                            self.nrot += 1
                            break

            if (rot_present[2]&rot_present[3]&(rot_present[4]|rot_present[5])):
                break
        return

    def is_valid_op(self, symmop,verbose=True):
        """
        Check if a particular symmetry operation is a valid symmetry operation
        for a molecule, i.e., the operation maps all atoms to another
        equivalent atom.

        Args:
            symmop (SymmOp): Symmetry operation to test.

        Returns:
            (bool): Whether SymmOp is valid for Molecule.
        """
        
        distances = []
        mol0 = self.mol.copy()
        mol1 = self.mol.copy()
        mol1.positions = mol1.positions.dot(symmop)
        workmol = mol0+mol1
        other_indices = list(range(len(mol0),len(workmol),1))
        for atom_index in range(0,len(mol0),1):
            dist = workmol.get_distances(atom_index,other_indices,mic=False)
            distances.append(np.amin(dist))
        distances = np.array(distances)
        return (distances<self.tol).all()





#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import os
import sys
import numpy
import ase
import logging
import scipy
import scipy.optimize

from collections import defaultdict


from autografs.utils.sbu        import read_sbu_database
from autografs.utils.sbu        import SBU
from autografs.utils.topology   import read_topologies_database
from autografs.utils.topology   import Topology
from autografs.framework        import Framework

logger = logging.getLogger(__name__) 

class Autografs(object):
    """Framework maker class to generate ASE Atoms objects from topologies.

    AuToGraFS: Automatic Topological Generator for Framework Structures.
    Addicoat, M. a, Coupry, D. E., & Heine, T. (2014).
    The Journal of Physical Chemistry. A, 118(40), 9607–14. 
    """

    def  __init__(self):
        """Constructor for the Autografs framework maker.
        """
        logger.info("Reading the topology database.")
        self.topologies : dict = read_topologies_database()
        logger.info("Reading the building units database.")
        self.sbu        : dict = read_sbu_database()


    def make(self,
             topology_name : str,
             sbu_names     : list  = None,
             sbu_dict      : dict  = None,
             supercell     : tuple = (1,1,1)) -> ase.Atoms :
        """Create a framework using given topology and sbu.

        Main funtion of Autografs. The sbu names and topology's
        are to be taken from the compiled databases. The sbu_dict
        can also be passed for multiple components frameworks.
        If the sbu_names is a list of tuples in the shape 
        (name,n), the number n will be used as a drawing probability
        when multiple options are available for the same shape.
        topology_name -- name of the topology to use
        sbu_names     -- list of names of the sbu to use
        sbu_dict -- (optional) one to one sbu to slot correspondance
                    in the shape {index of slot : 'name of sbu'}
        """
        # ase.visualize.view(self.topologies[topology_name])
        logger.info("Starting the MOF generation.")
        # make the supercell prior to alignment
        if isinstance(supercell,int):
            supercell = (supercell,supercell,supercell)
        topology_atoms  = self.topologies[topology_name]
        ase.visualize.view(topology_atoms)
        if supercell!=(1,1,1):
            logger.info("{0}x{1}x{2} supercell of the topology is used.".format(*supercell))
            topology_atoms *= supercell
        # make the Topology object
        logger.info("Analysis of the topology.")
        topology = Topology(name  = topology_name,
                            atoms = topology_atoms)
        # container for the aligned SBUs
        aligned  = Framework()
        aligned.set_topology(topology=topology_atoms)
        alpha    = 0.0
        # identify the corresponding SBU
        logger.info("Scheduling the SBU to slot alignment.")
        try:
            if sbu_dict is None and sbu_names is not None:
                sbu_dict = self.get_sbu_dict(topology=topology,
                                             sbu_names=sbu_names)
            elif sbu_dict is not None:
                # the sbu_dict has been passed. if not SBU object, create them
                for k,v in sbu_dict.items():
                    if not isinstance(v,SBU):
                        assert isinstance(v,ase.Atoms)
                        if "name" in v.info.keys():
                            name = v.info["name"]
                        else:
                            name = str(k)
                        sbu_dict[k] = SBU(name=name,atoms=v)
            else:
                raise RuntimeError("Either supply sbu_names or sbu_dict.")
        except RuntimeError as exc:
            logger.error("Slot to SBU mappping interrupted.")
            logger.error("{exc}".format(exc))
        # some logging
        self.log_sbu_dict(sbu_dict=sbu_dict)
        # carry on
        for idx,sbu in sbu_dict.items():
            logger.debug("Treating slot number {idx}".format(idx=idx))
            logger.debug("\t|-->Aligning SBU {name}".format(name=sbu.name))
            # now align and get the scaling factor
            sbu,f = self.align(fragment=topology.fragments[idx],
                                     sbu=sbu)
            alpha += f
            aligned.append(index=idx,sbu=sbu)
        # refine the cell scaling using a good starting point
        aligned.refine(alpha0=alpha)
        return aligned

    def log_sbu_dict(self,
                     sbu_dict : dict = None) -> None:
        """Does some logging on the chosen SBU mapping."""
        for idx,sbu in sbu_dict.items():
            s00 = topology.shapes[idx][0]
            s01 = topology.shapes[idx][1]
            s10 = sbu.shape[0]
            s11 = sbu.shape[1]
            logging.info("Slot {sl}, {s00} {s01}".format(sl=idx,
                                                         s00=s00,
                                                         s01=s01))
            logging.info("\t|-->SBU {sbn} {s10} {s11}.".format(sbn=sbu.name,
                                                                s10=s10,
                                                                s11=s11))
        return None

    def get_topology(self, 
                     topology_name : str = None) -> object:
        """Generates and return a Topology object"""
        topology_atoms = self.topologies[topology_name]
        return Topology(name=topology_name, atoms=topology_atoms)

    def get_sbu_dict(self,
                     topology  : object,
                     sbu_names : list) -> dict:
        """Return a dictionary of SBU by corresponding fragment.

        This stage get a one to one correspondance between
        each topology slot and an available SBU from the list of names.
        TODO: For now, we take the first available if more are given,
        but we should be able to pass this directly to the class,
        or a dictionary of probabilities for the different SBU.
        We also need to implement a check on symmetry operators,
        to catch stuff like 'squares cannot fit in a rectangle slot'.
        topology  -- the Topology object
        sbu_names -- the list of SBU names as strings
        """
        logger.debug("Generating slot to SBU map.")
        weights  = defaultdict(list)
        by_shape = defaultdict(list)
        for name in sbu_names:
            # check if probabilities included
            if isinstance(name,tuple):
                name,p = name
                p    = float(p)
                name = str(name)
            else:
                p = 1.0
            # create the SBU object
            sbu = SBU(name=name,atoms=self.sbu[name])
            comp, slot = topology.has_compatible_slot(sbu=sbu)
            if not comp:
                continue
            weights[slot].append(p)
            by_shape[slot].append(sbu)
        # now fill the choices
        sbu_dict = {}
        for index,shape in topology.shapes.items():       
            # here, should accept weights also
            p = weights[shape]
            p /= numpy.sum(p)
            print(shape)
            print(p)
            sbu_chosen = numpy.random.choice(by_shape[shape],
                                                  p=p).copy()
            logger.debug("Slot {sl}: {sb} chosen with p={p}.".format(sl=index,
                                                                     sb=sbu_chosen.name,
                                                                     p=p))
            sbu_dict[index] = sbu_chosen
        return sbu_dict

    def align(self,
              fragment : ase.Atoms,
              sbu      : object) -> (ase.Atoms, float):
        """Return an aligned SBU.

        The SBU is rotated on top of the fragment
        using the procrustes library within scipy.
        a scaling factor is also calculated for all three
        cell vectors.
        fragment -- the slot in the topology, ASE Atoms
        sbu      -- object to align, ASE Atoms
        """
        # first, we work with copies
        fragment       = fragment.copy()
        # normalize and center
        fragment_cop         = fragment.positions.mean(axis=0)
        fragment.positions  -= fragment_cop
        sbu.atoms.positions -= sbu.atoms.positions.mean(axis=0)
        # identify dummies in sbu
        sbu_Xis = [x.index for x in sbu.atoms if x.symbol=="X"]
        # get the scaling factor
        size_sbu      = numpy.linalg.norm(sbu.atoms[sbu_Xis].positions,axis=1)
        size_fragment = numpy.linalg.norm(fragment.positions,axis=1)
        alpha         = numpy.mean(size_sbu/size_fragment)
        # TODO check initial scaling: it goes up too much with unit cell
        ncop = numpy.linalg.norm(fragment_cop)
        if ncop<1e-6:
            direction  = numpy.ones(3,dtype=numpy.float32)
            direction /= numpy.linalg.norm(direction)
        else:
            direction = fragment_cop / ncop
        # scaling for better alignment
        fragment.positions = fragment.positions.dot(numpy.eye(3)*alpha)
        alpha *= direction
        # getting the rotation matrix
        X0  = sbu.atoms[sbu_Xis].get_positions()
        X1  = fragment.get_positions()
        if X0.shape[0]>5:
            X0 = self.get_vector_space(X0)
            X1 = self.get_vector_space(X1)
        R,s = scipy.linalg.orthogonal_procrustes(X0,X1)
        sbu.atoms.positions = sbu.atoms.positions.dot(R)+fragment_cop
        fragment.positions += fragment_cop
        res_d = ase.geometry.distance(sbu.atoms[sbu_Xis],fragment)
        logger.debug("Residual distance: {d}".format(d=res_d))
        # tag the atoms
        sbu.transfer_tags(fragment)
        return sbu,alpha

    def get_vector_space(self,
                         X : numpy.ndarray) -> numpy.ndarray:
        """Returns a vector space as three points."""
        x0 = X[0]
        dots = [x0.dot(x)for x in X]
        x1 = X[numpy.argmin(dots)]
        # dots = [x1.dot(x)for x in X[1:]]
        # x2 = X[numpy.argmin(dots)]
        x2 = numpy.cross(x0,x1)
        return numpy.asarray([x0,x1,x2])

    def list_available_topologies(self,
                                  sbu_names  : list = [],
                                  full       : bool = True) -> list:
        """Return a list of topologies compatible with the SBUs

        For each sbu in the list given in input, refines first by coordination
        then by shapes within the topology. Thus, we do not need to analyze
        every topology.
        sbu  -- list of sbu names
        full -- wether the topology is entirely represented by the sbu
        """
        if sbu_names:
            logger.info("Checking topology compatibility.")
            topologies = []
            sbu = [SBU(name=n,atoms=self.sbu[n]) for n in sbu_names]
            for tk,tv in self.topologies.items():
                try:
                    topology = Topology(name=tk,atoms=tv)
                except Exception as exc:
                    logger.error("Topology {tk} not loaded: {exc}".format(tk=tk,exc=exc))
                    continue
                filled = {shape:False for shape in topology.get_unique_shapes()}
                allsbu = True
                comps  = [topology.has_compatible_slot(s) for s in sbu]
                for comp,shape in comps:
                    filled[shape] = comp
                    allsbu = (comp&allsbu)
                if full and allsbu and all(filled.values()):
                    logger.info("\tTopology {tk} fully available.".format(tk=tk))
                    topologies.append(tk)
                elif allsbu and not full:
                    logger.info("\tTopology {tk} partially available.".format(tk=tk))
                    topologies.append(tk)
        else:
            logger.info("Listing full database of topologies.")
            topologies = list(self.topologies.keys())
        return topologies

    def list_available_sbu(self,
                           topology_name : str  = None) -> dict:
        """Return the dictionary of compatible SBU.
        
        Filters the existing SBU by shape until only
        those compatible with a slot within the topology are left.
        TODO: use the symmetry operators instead of the shape itself.
        topology -- name of the topology in the database
        """
        av_sbu = defaultdict(list)
        if topology_name is not None:
            logger.info("Checking SBU compatibility.")
            topology = Topology(name=topology_name,
                                atoms=self.topologies[topology_name])
            topops = topology.get_unique_operations()
            shapes = topology.get_unique_shapes()
            print(shapes)
            # filter according to coordination first
            for sbuk,sbuv in self.sbu.items():
                c = len([x for x in sbuv if x.symbol=="X"])
                for shape in shapes:
                    # first condition is coordination
                    if c==shape[1]:
                        # now check symmops
                        try:
                            sbu = SBU(name=sbuk,
                                      atoms=sbuv)
                        except Exception as exc:
                            log.error("SBU {k} not loaded: {exc}".format(k=sbuk,exc=exc))
                        if sbu.shape == shape:
                            logger.info("SBU {k} is compatible by shape.".format(k=sbuk))
                            av_sbu[shape].append(sbuk)
                        elif sbu.is_compatible(topops[shape]):
                            logger.info("SBU {k} is compatible by symmetry.".format(k=sbuk))
                            av_sbu[shape].append(sbuk)
        else:
            logger.info("Listing full database of SBU.")
            av_sbu = list(self.sbu.keys())
        return dict(av_sbu)



if __name__ == "__main__":

    molgen         = Autografs()
    sbu_names      = ["Benzene_linear","Zn_mof5_octahedral"]
    topology_name  = "pcu"
    mof = molgen.make(topology_name=topology_name,sbu_names=sbu_names)
    atoms,_,_ = mof.get_atoms()
    ase.visualize.view(atoms)




#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

"""
AuToGraFS: Automatic Topological Generator for Framework Structures.
Addicoat, M. a, Coupry, D. E., & Heine, T. (2014).
The Journal of Physical Chemistry. A, 118(40), 9607–14. 
"""

__all__     = ["autografs","framework","utils"]
__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

from autografs.autografs import Autografs
from autografs.framework import Framework

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.basicConfig(format='%(asctime)s in %(name)s %(levelname)s: %(message)s', 
	                level=logging.INFO, 
	                datefmt='%I:%M:%S')



#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import os
import sys
import numpy
import _pickle as pickle

import ase
from ase              import Atom, Atoms
from ase.spacegroup   import crystal
from ase.data         import chemical_symbols
from ase.neighborlist import NeighborList


from scipy.cluster.hierarchy import fclusterdata as cluster

import warnings

from autografs.utils.pointgroup import PointGroup
from autografs.utils           import __data__

def read_cgd() -> dict:
    """Return a dictionary of topologies as ASE Atoms objects
    
    The format CGD is used mainly by the Systre software
    and by Autografs. All details can be read on the website
    http://rcsr.anu.edu.au/systre
    """
    root = os.path.join(__data__,"topologies")
    topologies = {}
    # we need the names of the groups and their correspondance in ASE spacegroup data
    # this was compiled using Levenshtein distances and regular expressions
    groups_file = os.path.join(root,"HermannMauguin.dat")
    grpf = open(groups_file,"rb")
    groups = {l.split()[0]:l.split()[1] for l in grpf.read().decode("utf8").splitlines()}
    grpf.close()
    # read the rcsr topology data
    topology_file = os.path.join(root,"nets.cgd")
    # the script as such starts here
    with open(topology_file,"rb") as tpf:
        text = tpf.read().decode("utf8")
        # split the file by topology
        topologies_raw = [t.strip().strip("CRYSTAL") for t in text.split("END")]
        topologies_len = len(topologies_raw)
        print("{0:<5} topologies before treatment".format(topologies_len))
        for topology_raw in topologies_raw:
            # read from the template.
            # the edges are easier to comprehend by edge center
            try:
                lines = topology_raw.splitlines()
                lines = [l.split() for l in lines if len(l)>2]
                name = None
                group = None
                cell = []
                symbols = []
                nodes = []
                for l in lines:
                    if l[0].startswith("NAME"):
                        name = l[1].strip()
                    elif l[0].startswith("GROUP"):
                        group = l[1]
                    elif l[0].startswith("CELL"):
                        cell = numpy.array(l[1:], dtype=numpy.float32)
                    elif l[0].startswith("NODE"):
                        this_symbol = chemical_symbols[int(l[2])]
                        this_node = numpy.array(l[3:], dtype=numpy.float32)
                        nodes.append(this_node)
                        symbols.append(this_symbol)
                    elif (l[0].startswith("#") and l[1].startswith("EDGE_CENTER")):
                        # linear connector
                        this_node = numpy.array(l[2:], dtype=numpy.float32)
                        nodes.append(this_node)
                        symbols.append("He")
                    elif l[0].startswith("EDGE"):
                        # now we append some dummies
                        s    = int((len(l)-1)/2)
                        midl = int((len(l)+1)/2)
                        x0  = numpy.array(l[1:midl],dtype=numpy.float32).reshape(-1,1)
                        x1  = numpy.array(l[midl:] ,dtype=numpy.float32).reshape(-1,1)
                        xx  = numpy.concatenate([x0,x1],axis=1).T
                        com = xx.mean(axis=0)
                        xx -= com
                        xx  = xx.dot(numpy.eye(s)*0.5)
                        xx += com
                        nodes   += [xx[0],xx[1]]
                        symbols += ["X","X"]
                nodes = numpy.array(nodes)
                if len(cell)==3:
                    # 2D net, only one angle and two vectors.
                    # need to be completed up to 6 parameters
                    pbc  = [True,True,False] 
                    cell = numpy.array(list(cell[0:2])+[10.0,90.0,90.0]+list(cell[2:]), dtype=numpy.float32)
                    # node coordinates also need to be padded
                    nodes = numpy.pad(nodes, ((0,0),(0,1)), 'constant', constant_values=0.0)
                elif len(cell)<3:
                    continue
                else:
                    pbc = True
                # now some postprocessing for the space groups
                setting = 1
                if ":" in group:
                    # setting might be 2
                    group, setting = group.split(":")
                    try: 
                        setting = int(setting.strip())
                    except ValueError:
                        setting = 1
                # ASE does not have all the spacegroups implemented yet
                if group not in groups.keys():
                    continue
                else:
                    # generate the crystal
                    group     = int(groups[group])
                    topology  = crystal(symbols=symbols,
                                        basis=nodes,
                                        spacegroup=group,
                                        setting=setting,
                                        cellpar=cell,
                                        pbc=pbc,
                                        primitive_cell=False,
                                        onduplicates="keep")
                    topologies[name] = topology
            except Exception as expt:
                continue
    return topologies


def read_sbu(path    : None = None,
             formats : list =["xyz"]) -> dict:
    """Return a dictionary of Atoms objects.

    If the path is not specified, use the default library.
    TODO: Should use a chained iterable of path soon.
    path    -- where to find the sbu
    formats -- what molecular file format to read
    """
    if path is not None:
        path = os.path.abspath(path)
    else:
        path = os.path.join(__data__,"sbu")
    SBUs = {}
    for sbu_file in os.listdir(path):
        ext = sbu_file.split(".")[-1]
        if ext in formats:
            for sbu in ase.io.iread(os.path.join(path,sbu_file)):
                try:
                    name  = sbu.info["name"]
                    SBUs[name] = sbu
                except Exception as e:
                    continue
    return SBUs


def write_gin(path,
              atoms,
              bonds,
              mmtypes):
    """Write an GULP input file to disc"""
    with open(path,"w") as fileobj:
        fileobj.write('opti conp molmec cartesian noautobond\n')
        pbc = atoms.get_pbc()
        if pbc.any():
            cell = atoms.get_cell().tolist()
            if not pbc[2]:
                fileobj.write('{0}\n'.format('svectors'))
                fileobj.write('{0:.3f} {1:.3f} {2:.3f}\n'.format(*cell[0]))
                fileobj.write('{0:.3f} {1:.3f} {2:.3f}\n'.format(*cell[1]))
            else:
                fileobj.write('{0}\n'.format('vectors'))
                fileobj.write('{0:.3f} {1:.3f} {2:.3f}\n'.format(*cell[0]))
                fileobj.write('{0:.3f} {1:.3f} {2:.3f}\n'.format(*cell[1]))
                fileobj.write('{0:.3f} {1:.3f} {2:.3f}\n'.format(*cell[2]))
        fileobj.write('{0}\n'.format('cartesian'))
        symbols = atoms.get_chemical_symbols()
        #We need to map MMtypes to numbers. We'll do it via a dictionary
        symb_types=[]
        mmdic={}
        types_seen=1
        for m,s in zip(mmtypes,symbols):
            if m not in mmdic:
                mmdic[m]="{0}{1}".format(s,types_seen)
                types_seen+=1
                symb_types.append(mmdic[m])
            else:
                symb_types.append(mmdic[m])
        # write it
        for s, (x, y, z), in zip(symb_types, atoms.get_positions()):
            fileobj.write('{0:<4} {1:<7} {2:<15.8f} {3:<15.8f} {4:<15.8f}\n'.format(s,'core',x,y,z))
        fileobj.write('\n')
        bondstring = {4:'quadruple',
                      3:'triple',
                      2:'double',
                      1.5:'resonant',
                      1.0:'',
                      0.5:'half',
                      0.25:'quarter'}
        #write the bonding
        for (i0,i1),b in numpy.ndenumerate(bonds):
            if i0<i1 and b>0.0:
                fileobj.write('{0} {1:<4} {2:<4} {3:<10}\n'.format('connect',i0+1,i1+1,bondstring[b]))
        fileobj.write('\n')
        fileobj.write('{0}\n'.format('species'))
        for  k,v in  mmdic.items():
            fileobj.write('{0:<5} {1:<5}\n'.format(v,k))
        fileobj.write('\n')
        fileobj.write('library uff4mof\n')
        fileobj.write('\n')
        name = ".".join(path.split("/")[-1].split(".")[:-1])
        fileobj.write('output movie xyz {0}.xyz\n'.format(name))
        if pbc.any():
            fileobj.write('output cif {0}.cif\n'.format(name))
        return None


#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import os

__all__     = ["pointgroup","topology","sbu","operations","mmanalysis","io"]
__data__    = os.path.join("/".join(os.path.dirname(__file__).split("/")[:-1]),"data")



#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import os
import sys
import numpy
import scipy
import typing
import ase
import itertools
import logging
from collections import defaultdict

from autografs.utils.sbu        import read_sbu_database
from autografs.utils.topology   import read_topologies_database
from autografs.utils.mmanalysis import analyze_mm 
from autografs.utils.topology   import Topology

logger = logging.getLogger(__name__) 


class Framework(object):
    """
    The Framework object contain the results of an Autografs run.

    Designed to store and post-process the aligned fragments and the
    corresponding topology of an Autografs-made structure. 
    Methods available are setters and getters for the topology, individual SBU,
    rescaling of the total structure, and finally the connecting and returning
    of a clean ASE Atoms object.
    The SBUs are assumed to already be aligned and tagged by the 
    Framework generator, and both bonds and mmtypes are assumed to be ordered.
    The bond matrices are also assumed to be block symmetrical.
    """

    def __init__(self,
                 topology : ase.Atoms = None,
                 SBU      : dict      = {},
                 mmtypes  : numpy.ndarray = [],
                 bonds    : numpy.ndarray = []) -> None:
        """Constructor for the Framework class.

        Can be initialized empty and filled consecutively.
        topology -- ASE Atoms object.
        SBU      -- list of ASE Atoms objects.
        mmtypes  -- array of type names. e.g: 'C_R','O_3'...
        bonds    -- block-symmetric matrix of bond orders.
        """
        logger.debug("Creating Framework instance.")
        self.topology = topology
        self.SBU      = SBU
        self.mmtypes  = numpy.asarray(mmtypes)
        self.bonds    = numpy.asarray(bonds)
        # keep a dict of elements to delete in 
        # each SBU at connection time. Necessary for example
        # during iterative functionalization.
        self._todel = defaultdict(list)
        return None

    def __contains__(self, 
                     obj : object) -> None:
        """Iterable intrinsic"""
        r = False
        if hasattr(obj, 'atoms'):
            r = any([obj.atoms==sbu.atoms for sbu in self.SBU.values()])
        return r

    def __delitem__(self,
                    key : str) -> None:
        """Indexable intrinsic"""
        del self.SBU[key]
        return None

    def __setitem__(self,
                    key : str,
                    obj : object) -> None:
        """Indexable intrinsic"""
        if hasattr(obj, 'atoms'):
            self.SBU[key] = object
        return None

    def __getitem__(self,
                    key : str) -> None:
        """Indexable intrinsic"""
        return self.SBU[key]

    def __len__(self):
        """Sizeable intrinsic"""
        return len(self.SBU)

    def __iter__(self):
        """Iterable intrinsic"""
        return iter(self.SBU.items())

    def copy(self) -> object:
        """Return a copy of itself as a new instance"""
        new = self.__class__(topology=self.get_topology(),
                             SBU = self.SBU.copy(),
                             mmtypes= self.get_mmtypes(),
                             bonds=self.get_bonds())
        new._todel = self._todel.copy()
        return new

    def set_topology(self,
                     topology : ase.Atoms) -> None:
        """Set the topology attribute with an ASE Atoms object."""
        logger.debug("Setting topology.")
        self.topology = topology.copy()
        return None

    def get_topology(self) -> ase.Atoms:
        """Return a copy of the topology attribute as an ASE Atoms object."""
        return self.topology.copy()

    def get_supercell(self,
                      m : tuple = (2,2,2)) -> object:
        """Return a framework supercell usin m as multiplier"""
        # TODO !!! [CRITICAL] correct tagging needed!!!.
        if isinstance(m,int):
            m = (m,m,m)
        logger.info("Creating supercell {0}x{1}x{2}.".format(*m))
        # get the offset direction ranges
        x = list(range(0,m[0],1))
        y = list(range(0,m[1],1))
        z = list(range(0,m[2],1)) 
        # new framework object
        supercell = self.copy()
        ocell = supercell.topology.get_cell()
        otopo = supercell.topology.copy()
        cellfactor = numpy.asarray([x[-1]+1,y[-1]+1,z[-1]+1],dtype=float)
        newcell = ocell.dot(numpy.eye(3)*cellfactor)
        supercell.topology.set_cell(newcell,scale_atoms=False)
        # we need a topology object to correctly tag.
        supertopo = Topology(name="supercell",atoms=otopo*m)
        noff  = 0
        L     = len(otopo)
        # iterate over offsets and add the corresponding objects
        for offset in itertools.product(x,y,z):
            # central cell, ignore
            if offset==(0,0,0):
                for atom in otopo.copy():
                    if atom.symbol=="X":
                        continue
                    if atom.index not in supercell.SBU.keys():
                        continue
                    # directly tranfer new tags
                    sbu = supercell[atom.index]
                    sbu.transfer_tags(supertopo.fragments[atom.index])           
            else:
                noff += 1 
                coffset = ocell.dot(offset)
                for atom in otopo.copy():
                    atom.position += coffset
                    supercell.topology.append(atom)
                    if atom.symbol=="X":
                        continue
                    newidx = len(supercell.topology)-1
                    # check that the SBU was not deleted before
                    if atom.index not in supercell.SBU.keys():
                        continue
                    sbu = supercell[atom.index].copy()
                    sbu.atoms.positions += coffset
                    sbu.transfer_tags(supertopo.fragments[newidx])
                    supercell.append(index = newidx,
                                     sbu   = sbu,
                                     update= False)
                    supercell._todel[newidx] = list(supercell._todel[atom.index])
        # redo the tagging
        return supercell

    def append(self,
               index   : int,
               sbu     : ase.Atoms,
               update  : bool = False) -> None:
        """Append all data releted to a building unit in the framework.

        This includes the ASE Atoms object, the bonding matrix, and the 
        molecular mechanics atomic types as numpy arrays. These three objects 
        are related through indexing: sbu[i] has a MM type mmtypes[i] and 
        a bonding array of bonds[i,:] or bonds[:,i]
        sbu     -- the Atoms object
        bonds   -- the bonds numpy array, of size len(sbu) by len(sbu).
        mmtypes -- the MM atomic types.
        """
        # first append the atoms object to the list of sbu
        logger.debug("Appending SBU {n} to framework.".format(n=sbu.name))
        self.SBU[index] = sbu
        if update:
            # make the bonds matrix with a new block
            self.bonds   = self.get_bonds()
            # append the atom types
            self.mmtypes = self.get_mmtypes()
        return None

    def get_bonds(self) -> numpy.ndarray:
        """Return and update the current bond matrix"""
        logger.debug("Updating framework bond matrix.")
        bonds = []
        for _,sbu in self:
            bonds.append(sbu.bonds)
        bonds = scipy.linalg.block_diag(*bonds)
        self.bonds = bonds
        return bonds

    def get_mmtypes(self) -> numpy.ndarray:
        """Return and update the current bond matrix"""
        logger.debug("Updating framework MM types.")
        mmtypes = []
        for _,sbu in self:
            mmtypes.append(sbu.mmtypes)
        mmtypes = numpy.hstack(mmtypes)
        self.mmtypes = mmtypes
        return numpy.copy(self.mmtypes)

    def scale(self,
              alpha  : numpy.ndarray) -> None:
        """Scale the building units positions by a factor alpha.

        This uses the correspondance between the atoms in the topology
        and the building units in the SBU list. Indeed, SBU[i] is centered on 
        topology[i]. By scaling the topology, we obtain a new center for the 
        sbu.
        alpha -- scaling factor
        """
        logger.debug("Scaling framework by {0:.3f}x{1:.3f}x{2:.3f}.".format(*alpha))
        # get the scaled cell, normalized
        I     = numpy.eye(3)*alpha
        cell  = self.topology.get_cell()
        ncell = numpy.linalg.norm(cell,axis=0)
        cell  = cell.dot(I*1.73/ncell)
        self.topology.set_cell(cell,scale_atoms=True)
        # then center the SBUs on this position
        for i,sbu in self:
            center = self.topology[i]
            cop    = sbu.atoms.positions.mean(axis=0)
            sbu.atoms.positions += center.position - cop
        return None

    def refine(self,
               alpha0 : numpy.ndarray = [1.0,1.0,1.0]) -> None:
        """Refine cell scaling to minimize distances between dummies.

        We already have tagged the corresponding dummies during alignment,
        so we just need to calculate the MSE of the distances between 
        identical tags in the complete structure
        alpha0 -- starting point of the scaling search algorithm
        """
        logger.info("Refining unit cell.")
        def MSE(x : float) -> float:
            """Return cost of scaling as MSE of distances."""
            # scale with this parameter
            x = x*alpha0
            self.scale(alpha=x)
            atoms,_,_    = self.get_atoms(dummies=True)
            tags         = atoms.get_tags()
            # reinitialize stuff
            self.scale(alpha=1.0/x)
            # find the pairs...
            pairs = [numpy.argwhere(tags==tag) for tag in set(tags) if tag>0]
            pairs =  numpy.asarray(pairs).reshape(-1,2)
            # ...and the distances
            d = [atoms.get_distance(i0,i1,mic=True) for i0,i1 in pairs]
            d = numpy.asarray(d)
            mse = numpy.mean(d**2)
            logger.info("\tScaling error = {e:>5.3f}".format(e=mse))
            return mse
        # first get an idea of the bounds.
        # minimum cell of a mof should be over 2.0 Ang.
        # and directions with no pbc should be 1.0
        alpha0[alpha0<1e-6] = 20.0
        low  = 0.1
        high = 1.5
        if numpy.amin(low*alpha0)<2.0:
            low   = 1.5/numpy.amin(alpha0)
            high *= 0.1/low
        # optimize
        result = scipy.optimize.minimize_scalar(fun    = MSE,
                                                bounds = (low,high),
                                                method = "Bounded",
                                                options= {"disp"  : True,
                                                          "xatol" : 1e-02})
        # scale with result
        alpha = result.x*alpha0
        logger.info("Best scaling achieved by {0:.3f}x{1:.3f}x{2:.3f}.".format(*alpha))
        self.scale(alpha=alpha)
        return None

    def rotate(self,
               index : int,
               angle : float) -> None:
        """Rotate the SBU at index around a Cinf symmetry axis"""
        if self[index].shape[1]==2:
            logger.info("Rotating {idx} by {a}.".format(idx=index,a=angle))
            axis = [x.position for x in self[index].atoms 
                               if x.symbol=="X"]
            axis = numpy.asarray(axis)
            axis = axis[0]-axis[1]
            self[index].atoms.rotate(v=axis,a=angle)
        return None

    def flip(self,
             index : int) -> None:
        """Flip the SBU at index around a C* symmetry axis or Sigmav plane"""
        logger.info("Flipping {idx}".format(idx=index))
        if self[index].shape[1]==2:
            axis = [x.position for x in self[index].atoms 
                               if x.symbol=="X"]
            axis = numpy.asarray(axis)
            axis = axis[0]-axis[1]
            self[index].atoms.rotate(v=axis,a=180.0)
        else:
            sigmav = [op[2] for op in self[index].symmops["sigma"]
                            if op[0]=="v"]
            if sigmav:
                cop = sbu.atoms.positions.mean(axis=0)
                pos = sbu.atoms.positions - cop
                pos = pos.dot(sigmav[0])  + cop
                self[index].atoms.set_positions(pos)
        return None

    def list_functionalizable_sites(self,symbol=None) -> list:
        """Return a list of tuple for functionalizable sites"""
        if symbol is not None:
            logger.info("Listing available functionalizable {s}".format(symbol))
        else:
            logger.info("Listing all available functionalization sites")
        sites = []
        for idx,sbu in self:
            bonds = sbu.bonds
            for atom in sbu.atoms:
                if symbol is not None and atom.symbol!=symbol:
                    continue
                if atom.symbol=="X":
                    continue
                bidx = numpy.where(bonds[atom.index,:]>0.0)[0]
                if len(bidx)!=1:
                    continue
                elif bonds[atom.index,bidx[0]]!=1.0:
                    continue
                else:
                    sites.append((idx,atom.index))
        return sites

    def functionalize(self,
                      where : tuple,
                      fg    : ase.Atoms) -> None:
        """Modify a valid slot atom to become a functional group.
        
        The specified slot index and atom index within the slot
        are examined: if the corresponding atom is a single
        atom connected by a bond order of 1 to exactly one other
        atom, it can be functionalized. The given functional group
        has to have exactly one dummy atom. For best performance,
        set the x-func distance at the C-H bond distance.
        where -- (slot index, atom index)
        fg    -- ASE Atoms to replace the atom in where by.
        TODO : move the SBU import to toplevel
        """
        from autografs.utils.sbu import SBU
        sidx, aidx = where
        # create the SBU object for the functional group
        fg_name = "func:{0}".format(fg.get_chemical_formula(mode="hill"))
        logger.info("Functionalization of atom {1} in slot {0}.".format(*where))
        try:
            logger.info("\t|--> replace by {f}.".format(fg_name))
            fg = SBU(name=fg_name,atoms=fg)
            # center the positions
            fg_cop = fg.atoms.positions.mean(axis=0)
            fg.atoms.positions -= fg_cop
            # check that only one dummy exists
            xidx = [x.index for x in fg.atoms if x.symbol=="X"]
            assert len(xidx)==1
            xidx  = xidx[0]
            # center the sbu
            sbu   = self.SBU[sidx].atoms
            sbu_name = self.SBU[sidx].name
            sbu_cop = sbu.positions.mean(axis=0)
            sbu.positions -= sbu_cop
            # find the bonds
            bonds = self.SBU[sidx].bonds
            bidx  = numpy.where(bonds[aidx]>0.0)[0]
            # check that only one bond exists
            assert len(bidx)==1 and bonds[aidx,bidx[0]]==1.0
            # now get vectors to align
            fgbidx = numpy.where(fg.bonds[xidx]>0.0)[0]
            assert len(fgbidx)==1 and fg.bonds[xidx,fgbidx]==1.0
            # func-x
            v0 = fg.atoms.positions[fgbidx]-fg.atoms.positions[xidx]
            # where[1]-bonded atom
            v1 = sbu.positions[aidx]-sbu.positions[bidx]
            R,s = scipy.linalg.orthogonal_procrustes(v0,v1)
            fg.atoms.positions = fg.atoms.positions.dot(R)
            fg.atoms.positions -= fg.atoms.positions[xidx]
            fg.atoms.positions += sbu.positions[bidx]
            # create the new object
            # keep note of what to delete.
            self._todel[sidx] += [aidx,xidx+len(sbu)]
            sbu += fg.atoms
            sbu.positions += sbu_cop
            self.SBU[sidx].set_atoms(sbu,analyze=False)
            self.SBU[sidx].bonds = scipy.linalg.block_diag(bonds,
                                                           fg.bonds)
            self.SBU[sidx].mmtypes = numpy.hstack([self.SBU[sidx].mmtypes,
                                                  fg.mmtypes])
        except Exception as exc:
            logger.error("\t|--> ERROR WHILE FUNCTIONALIZING.")
            logger.error("\t|--> {exc}".format(exc))
        return None

    def get_atoms(self,
                  dummies : bool = False) -> ase.Atoms:
        """Return the concatenated Atoms objects.

        The concatenation can either remove the dummies and
        connect the corresponding atoms or leave hem in place
        clean -- remove the dummies if True
        """
        logger.debug("Creating ASE Atoms from framework.")
        if dummies:
            logger.debug("\tDummies will be kept.")
            logger.debug("\tNo connection between SBU will occur.")
        else:
            logger.debug("\tDummies will be removed during connection.")
        # concatenate every sbu into one Atoms object
        framework = self.copy()
        cell = framework.topology.get_cell()
        pbc  = framework.topology.get_pbc()
        structure = ase.Atoms(cell=cell,pbc=pbc)
        for idx,sbu in framework:
            atoms = sbu.atoms
            todel = framework._todel[idx]
            if len(todel)>0:
                del atoms[todel]
            framework[idx].set_atoms(atoms,analyze=True)
            structure += atoms
        bonds   = framework.get_bonds()
        mmtypes = framework.get_mmtypes()
        symbols = numpy.asarray(structure.get_chemical_symbols())
        if not dummies:
            # keep track of dummies
            xis   = [x.index for x in structure if x.symbol=="X"]
            tags  = structure.get_tags()
            pairs = [numpy.argwhere(tags==tag) for tag in set(tags[xis])]
            # raise
            for pair in pairs:
                # if lone dummy, cap with hydrogen
                if len(pair)==1:
                    xi0 = pair[0]
                    xis.remove(xi0)
                    symbols[xi0] = "H"
                    mmtypes[xi0] = "H_" 
                else:
                    xi0,xi1 = pair
                    bonds0  = numpy.where(bonds[:,xi0]>0.0)[0]
                    bonds1  = numpy.where(bonds[:,xi1]>0.0)[0]
                    # dangling bit, mayhaps from defect
                    if len(bonds0)==0 and len(bonds1)!=0:
                        xis.remove(xi1)
                        symbols[xi1] = "H"
                        mmtypes[xi1] = "H_" 
                    elif len(bonds1)==0 and len(bonds0)!=0:
                        xis.remove(xi0)
                        symbols[xi0] = "H"
                        mmtypes[xi0] = "H_"                         
                    else:
                        # the bond order will be the maximum one
                        bo      = max(numpy.amax(bonds[xi0,:]),
                                      numpy.amax(bonds[xi1,:]))
                        # change the bonds
                        ix        = numpy.ix_(bonds0,bonds1)
                        bonds[ix] = bo
                        ix        = numpy.ix_(bonds1,bonds0)
                        bonds[ix] = bo
            # book keeping on what has disappeared
            structure.set_chemical_symbols(symbols)
            bonds   = numpy.delete(bonds,xis,axis=0)
            bonds   = numpy.delete(bonds,xis,axis=1)
            mmtypes = numpy.delete(mmtypes,xis)
            del structure[xis]
        return structure, bonds, mmtypes

    def write(self,
              f   : str = "./mof",
              ext : str = "gin") -> None:
        """Write a chemical information file to disk in selected format"""
        atoms,bonds,mmtypes = self.get_atoms(dummies=False)
        path = os.path.abspath("{path}.{ext}".format(path=f,ext=ext))
        logger.info("Framework saved to disk at {p}.".format(p=path))
        if ext=="gin":
            from autografs.utils.io import write_gin
            write_gin(path,atoms,bonds,mmtypes)
        else:
            ase.io.write(path,atoms)



#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import numpy

def inertia(xyz : numpy.ndarray,
            W   : numpy.ndarray) -> numpy.ndarray:
    """Return the inertia matrix of coordinates, weighted"""
    x = xyz[:,0]
    y = xyz[:,1]
    z = xyz[:,2]
    x2 = x*x
    y2 = y*y
    z2 = z*z
    Ixx =  numpy.sum((y2+z2)*W)
    Iyy =  numpy.sum((x2+z2)*W)
    Izz =  numpy.sum((x2+y2)*W)
    Ixy = -numpy.sum((x *y )*W)
    Iyz = -numpy.sum((y *z )*W)
    Ixz = -numpy.sum((x *z )*W)
    I = numpy.array([[Ixx,Ixy,Ixz],
                     [Ixy,Iyy,Iyz],
                     [Ixz,Iyz,Izz]],dtype=numpy.float32)
    return I
    
def rotation(axis  : numpy.ndarray,
             order : float) -> numpy.ndarray:
    """Return a rotation matrix around the axis"""
    M = numpy.eye(3)
    axis /= numpy.linalg.norm(axis)
    v0      = axis[0]
    v1      = axis[1]
    v2      = axis[2]
    theta   = 2.0*numpy.pi/order
    costh   = numpy.cos(theta)
    sinth   = numpy.sin(theta)
    M[0,0] = costh + (1.0-costh)*v0**2
    M[1,1] = costh + (1.0-costh)*v1**2
    M[2,2] = costh + (1.0-costh)*v2**2
    M[1,0] = (1.0-costh)*v0*v1 + v2*sinth
    M[0,1] = (1.0-costh)*v0*v1 - v2*sinth
    M[2,0] = (1.0-costh)*v0*v2 - v1*sinth
    M[0,2] = (1.0-costh)*v0*v2 + v1*sinth
    M[2,1] = (1.0-costh)*v1*v2 + v0*sinth
    M[1,2] = (1.0-costh)*v1*v2 - v0*sinth   
    return M
    
    
def reflection(axis : numpy.ndarray) -> numpy.ndarray:
    """Return a reflection matrix around the axis"""
    M = numpy.eye(3)
    axis /= numpy.linalg.norm(axis)
    v0      = axis[0]
    v1      = axis[1]
    v2      = axis[2]
    M[0,0] = 1.0-2.0*v0*v0
    M[1,1] = 1.0-2.0*v1*v1
    M[2,2] = 1.0-2.0*v2*v2
    M[1,0] =    -2.0*v0*v1 
    M[0,1] =    -2.0*v0*v1 
    M[2,0] =    -2.0*v0*v2 
    M[0,2] =    -2.0*v0*v2 
    M[2,1] =    -2.0*v1*v2 
    M[1,2] =    -2.0*v1*v2
    return M


def procrustes(X : numpy.ndarray,
               Y : numpy.ndarray,
               method : str = "Q") -> (numpy.ndarray,float):
    """Return the optimal rotation between two sets of coordinates"""
    X = numpy.array(X,dtype=float)
    Y = numpy.array(Y,dtype=float)
    if method == "Q":
        #covariance matrix
        H = numpy.dot(X.T, Y)
        # four dimensional matrix for quaternion calculation
        P = numpy.array([[H[0,0]+H[1,1]+H[2,2], H[1,2]-H[2,1], H[2,0]-H[0,2], H[0,1]-H[1,0]],
                         [H[1,2]-H[2,1], H[0,0]-H[1,1]-H[2,2], H[0,1]+H[1,0], H[2,0]+H[0,2]],
                         [H[2,0]-H[0,2], H[0,1]+H[1,0], H[1,1]-H[0,0]-H[2,2], H[1,2]+H[2,1]],
                         [H[0,1]-H[1,0], H[2,0]+H[0,2], H[1,2]+H[2,1], H[2,2]-H[1,1]-H[0,0]]])
        # compute eigenvalues
        eigenvalues, eigenvectors = numpy.linalg.eigh(P)
        # quaternion
        q = eigenvectors[:, numpy.argmax(eigenvalues)]
        q /= numpy.linalg.norm(q)
        #orthogonal rotation matrix
        R = numpy.array([[(q[0]**2)+(q[1]**2)-(q[2]**2)-(q[3]**2), 2*(q[1]*q[2]-q[0]*q[3]), 2*(q[1]*q[3]+q[0]*q[2])],
                         [2*(q[1]*q[2]+q[0]*q[3]), (q[0]**2)-(q[1]**2)+(q[2]**2)-(q[3]**2), 2*(q[2]*q[3]-q[0]*q[1])],
                         [2*(q[1]*q[3]-q[0]*q[2]), 2*(q[2]*q[3]+q[0]*q[1]), (q[0]**2)-(q[1]**2)-(q[2]**2)+(q[3]**2)]])
        scale = 1.0
    elif method == "SVD":
        u, w, vt = numpy.linalg.svd(X.T.dot(Y).T)
        R = u.dot(vt)
        scale = w.sum()
    else:
        raise NotImplementedError("Unknown method. Implemented are SVD or Q")
    return R, scale


#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import numpy
import os
from scipy.sparse import csgraph

from ase.data import covalent_radii
from ase.neighborlist import NeighborList
from ase import Atom
import ase
from itertools import combinations
from ase.visualize import view

from autografs.utils import __data__

def is_metal(symbols : list) -> bool:
    """Check wether symbols in a list are metals"""
    symbols = numpy.array([symbols]).flatten()
    metals = ['Li','Be','Al','Sc','Ti','V','Cr','Mn','Fe','Co','Ni','Cu','Zn',
              'Ga','Ge','Y','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In',
              'Sn','Sb','La','Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb','Dy','Ho',
              'Er','Tm','Yb','Lu','Hf','Ta','W','Re','Os','Ir','Pt','Au','Hg',
              'Tl','Pb','Bi','Po','Fr','Ra','Ac','Th','Pa','U','Np','Pu','Am',
              'Cm','Bk','Cf','Es','Fm','Md','No','Lr','Rf','Db','Sg','Bh','Hs',
              'Mt','Ds','Rg','Cn','Nh','Fl','Mc','Lv','Ts','Og']
    return numpy.isin(symbols,metals)

def is_alkali(symbols : list) -> bool:
    """Check wether symbols in a list are alkali"""
    symbols = numpy.array([symbols]).flatten()
    alkali = ['Li','Be','Na','Mg','K','Ca','Rb','Sr','Cs','Ba']
    return numpy.isin(symbols,alkali)


def read_uff_library(library : str = "uff4mof") -> numpy.ndarray:
    """Return the UFF library as a numpy array"""
    uff_file = os.path.join(__data__,"uff/{0}.csv".format(library))
    with open(uff_file,"r") as lib:
        lines = [l.split(",") for l in lib.read().splitlines() if not l.startswith("#")]
        #symbol,radius,angle,coordination
        ufflib   = {s:numpy.array([r,a,c],dtype=numpy.float32) for s,r,a,c in lines}
    return ufflib

def get_bond_matrix(sbu : ase.Atoms) -> numpy.ndarray:
    """Guesses the bond order in neighbourlist based on covalent radii 
    the radii for BO > 1 are extrapolated by removing 0.1 Angstroms by order 
    see Beatriz Cordero, Veronica Gomez, Ana E. Platero-Prats, Marc Reves, Jorge Echeverria,
    Eduard Cremades, Flavia Barragan and Santiago Alvarez (2008). 
    "Covalent radii revisited". 
    Dalton Trans. (21): 2832-2838 
    http://dx.doi.org/10.1039/b801115j
    """
    # first guess
    bonds     = numpy.zeros((len(sbu),len(sbu)))
    symbols   = numpy.array(sbu.get_chemical_symbols())
    numbers   = numpy.array(sbu.get_atomic_numbers())
    positions = numpy.array(sbu.get_positions())
    BO1 = numpy.array([covalent_radii[n] if n>0 else 0.35 for n in numbers])
    BO2 = BO1 - 0.15
    BO3 = BO2 - 0.15
    nl1 = NeighborList(cutoffs=BO1,bothways=True,self_interaction=False,skin=0.1)
    nl2 = NeighborList(cutoffs=BO2,bothways=True,self_interaction=False,skin=0.1)
    nl3 = NeighborList(cutoffs=BO3,bothways=True,self_interaction=False,skin=0.1)
    nl1.build(sbu); nl2.build(sbu); nl3.build(sbu)
    for atom in sbu:
        i1, _ = nl1.get_neighbors(atom.index)
        i2, _ = nl2.get_neighbors(atom.index)
        i3, _ = nl3.get_neighbors(atom.index)
        bonds[atom.index,i1] = 1.0
        bonds[atom.index,i2] = 2.0
        bonds[atom.index,i3] = 3.0
    # cleanup with particular cases
    # identify particular atoms
    hydrogens = numpy.where(symbols=="H")[0]
    metals    = numpy.where(is_metal(symbols))[0]
    alkali    = numpy.where(is_alkali(symbols))[0]
    # the rest is dubbed "organic"
    organic   = numpy.ones(bonds.shape)
    organic[hydrogens,:] = False
    organic[metals   ,:] = False
    organic[alkali   ,:] = False
    organic[:,hydrogens] = False
    organic[:,metals   ] = False
    organic[:,alkali   ] = False    
    organic = numpy.where(organic)[0]
    # Hydrogen has BO of 1
    bonds_h = bonds[hydrogens]
    bonds_h[bonds_h>1.0] = 1.0
    bonds[hydrogens,:] = bonds_h
    bonds[:,hydrogens] = bonds_h.T
    #Metal-Metal bonds: if no special case, nominal bond
    ix  = numpy.ix_(metals,metals)
    bix = bonds[ix]
    bix[numpy.nonzero(bix)] = 0.25
    bonds[ix] = bix
    # no H-Metal bonds
    ix = numpy.ix_(metals,hydrogens)
    bonds[ix] = 0.0
    ix = numpy.ix_(hydrogens,metals)
    bonds[ix] = 0.0
    # no alkali-alkali bonds
    ix = numpy.ix_(alkali,alkali)
    bonds[ix] = 0.0
    # no alkali-metal bonds
    ix = numpy.ix_(metals,alkali)
    bonds[ix] = 0.0
    ix = numpy.ix_(alkali,metals)
    bonds[ix] = 0.0    
    # metal-organic is coordination bond
    ix  = numpy.ix_(metals,organic)
    bix = bonds[ix]
    bix[numpy.nonzero(bix)] = 0.5
    bonds[ix] = bix
    ix = numpy.ix_(organic,metals)
    bix = bonds[ix]
    bix[numpy.nonzero(bix)] = 0.5
    bonds[ix] = bix
    # aromaticity and rings
    rings = []
    # first, use the compressed sparse graph object
    # we only care about organic bonds and not hydrogens
    graph_bonds = numpy.array(bonds>=1.0,dtype=float)
    graph_bonds[hydrogens,:] = 0.0
    graph_bonds[:,hydrogens] = 0.0
    graph = csgraph.csgraph_from_dense(graph_bonds)
    for sg in graph.indices:
        subgraph = graph[sg]
        for i,j in combinations(subgraph.indices,2):
            t0 = csgraph.breadth_first_tree(graph, i_start=i, directed=False)
            t1 = csgraph.breadth_first_tree(graph, i_start=j, directed=False)
            t0i = t0.indices
            t1i = t1.indices
            ring = sorted(set(list(t0i)+list(t1i)+[i,j,sg]))
            # some conditions
            seen        = (ring in rings)
            isring      = (sorted(t0i[1:])==sorted(t1i[1:]))
            bigenough   = (len(ring)>=5)
            smallenough = (len(ring)<=10)
            if isring and not seen and bigenough and smallenough:
                rings.append(ring)
    # we now have a list of all the shortest rings within 
    # the molecular graph. If planar, the ring might be aromatic
    aromatic_epsilon = 0.1
    aromatic = []
    for ring in rings:
        homocycle   = (symbols[ring]=="C").all()
        heterocycle = numpy.in1d(symbols[ring],numpy.array(["C","S","N","O"])).all()
        if (homocycle and (len(ring)%2)==0) or heterocycle:
            ring_positions = positions[ring]
            # small function for coplanarity
            coplanar = all([numpy.linalg.det(numpy.array(x[:3])-x[3])<aromatic_epsilon
                            for x in combinations(ring_positions,4)])
            if coplanar:
                aromatic.append(ring)
    # aromatic bond fixing
    aromatic  = numpy.array(aromatic).ravel()
    ix        = numpy.ix_(aromatic,aromatic)
    bix       = bonds[ix]
    bix[numpy.nonzero(bix)] = 1.5
    bonds[ix] = bix
    # hydrogen bonds
    # TODO
    return bonds

def uff_symbol(atom : ase.Atoms) -> str:
    """Returns the first twol letters of a UFF parameters"""
    sym  = atom.symbol
    if len(sym) == 1 :
        sym =''.join([sym,'_']) 
    return sym

def best_angle(a       : int,
               sbu     : ase.Atoms,
               indices : numpy.ndarray) -> float:
    """Calculates the most common angle around an atom"""
    # linear case
    if len(indices)<=1:
        da = 180.0
    else:
        angles = numpy.array([sbu.get_angle(a1,a,a3,mic=True) 
                              for a1,a3 in combinations(indices,2)]).reshape(-1,1)
        if angles.shape[0]>1:
            # do some clustering on the angles, keep most frequent
            from scipy.cluster.hierarchy import fclusterdata as cluster
            clusters = cluster(angles, 10.0, criterion='distance')
            counts   = numpy.bincount(clusters)
            da = angles[numpy.where(clusters==numpy.argmax(counts))].mean()
        else:
            da = angles[0,0]
    return da

def best_radius(a       : int,
                sbu     : ase.Atoms,
                indices : numpy.ndarray,
                ufflib  : numpy.ndarray) -> float:
    """Return the radius, according to the neighbors of an atom"""
    if len(indices)==0:
        d1 = 0.7
    else:
        # the average of covalent radii will be used for distances
        others  = [uff_symbol(at) for at in sbu[indices] if at.symbol!="X"]
        if len(others)==0:
            d1 = 0.7
        else:
            d1      = numpy.array([numpy.array([v[0] for k,v in ufflib.items() 
                                   if k.startswith(s)]).mean() for s in others]).mean()
    # get the distances also
    d0  = sbu.get_distances(a,indices,mic=True).mean()
    dx  = d0-d1 
    return dx

def best_type(dx     : float,
              da     : float,
              dc     : int  ,
              ufflib : numpy.ndarray,
              types  : list) -> str:
    """Chooses the best UFF type according to neighborhood."""
    mincost = 1000.0
    mintyp  = None 
    for typ in types:
        xx,aa,cc = ufflib[typ]
        cost_x = ((dx-xx)**2)/2.50
        cost_a = ((da-aa)**2)/180.0
        cost_c = ((dc-cc)**2)/4.0
        cost   =  cost_x+cost_a+cost_c
        if cost<mincost:
            mintyp = typ
            mincost= cost
    return mintyp

def analyze_mm(sbu : ase.Atoms) -> (numpy.ndarray,numpy.ndarray):
    """Returns the UFF types and bond matrix for an ASE Atoms."""
    ufflib  = read_uff_library(library="uff4mof")
    bonds   = get_bond_matrix(sbu)
    mmtypes = [None,]*len(sbu)
    for atom in sbu:
        if atom.symbol == "X":
            continue
        # get the starting symbol in uff nomenclature
        symbol = uff_symbol(atom)
        # narrow the choices
        uff_types   = [k for k in ufflib.keys() if k.startswith(symbol)]
        these_bonds = bonds[atom.index].copy()
        # if only one choice, use it
        if len(uff_types)==1:
            mmtypes[atom.index] = uff_types[0]
        # aromatics are easy also
        elif (numpy.abs(these_bonds-1.5)<1e-6).any():
            uff_types = [typ for typ in uff_types if typ.endswith("R")]
            mmtypes[atom.index] = uff_types[0]
        else:
            indices = numpy.where(these_bonds>=0.25)[0]
            # coordination
            dc = len(indices)
            # angle
            da = best_angle(atom.index,sbu,indices)
            # radius
            dx = best_radius(atom.index,sbu,indices,ufflib)
            # complete data
            mmtypes[atom.index] = best_type(dx,da,dc,ufflib,uff_types)
    # now correct the dummies
    for xi in [x.index for x in sbu if x.symbol=="X"]:
        bonded = numpy.argmax(bonds[xi])
        mmtypes[xi] = mmtypes[bonded]
    mmtypes = numpy.array(mmtypes)
    bonds   = numpy.array(bonds)
    return bonds,mmtypes





#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
# Copyright : see accompanying license files for details

__author__  = "Damien Coupry"
__credits__ = ["Prof. Matthew Addicoat"]
__license__ = "MIT"
__maintainer__ = "Damien Coupry"
__version__ = '2.0.3'
__status__  = "beta"

import os
import sys
import numpy
import logging
import _pickle as pickle

import ase
from ase.spacegroup   import crystal
from ase.visualize    import view
from ase.data         import chemical_symbols
from ase.neighborlist import NeighborList

from collections import Counter

from scipy.cluster.hierarchy import fclusterdata as cluster
import warnings


from autografs.utils.pointgroup import PointGroup
from autografs.utils.mmanalysis import analyze_mm 
from autografs.utils            import __data__

logger = logging.getLogger(__name__) 


class SBU(object):
    """Container class for a building unit information"""

    def __init__(self,
                 name  : str,
                 atoms : ase.Atoms = None) -> None:
        """Constructor for a building unit, from an ASE Atoms."""
        logger.debug("New instance of SBU {0}".format(name))
        self.name  = name
        self.atoms = atoms 
        self.symmops = {}
        self.mmtypes = []
        self.bonds   = []
        self.shape = ("C1",0)
        if self.atoms is not None:
            self._analyze()
        logger.debug("{self}".format(self=self.__str__()))
        return None

    def __repr__(self) -> str:
        """Return representation full printing"""
        strings = []
        strings.append("\nSBU: {name}\n".format(name=self.name))
        # catch for empty object
        if self.atoms is None:
            return "".join(strings)
        strings.append("Point group : {sc}\n".format(sc=self.shape[0]))
        strings.append("Coordination: {co}\n".format(co=self.shape[1]))
        for atom in self.atoms:
            p0,p1,p2 = atom.position
            sy = atom.symbol
            tp = self.mmtypes[atom.index]
            s = "{sy:<3} {p0:>5.2f} {p1:>5.2f} {p2:>5.2f} {tp:<5}\n".format(sy=sy,
                                                                            p0=p0,
                                                                            p1=p1,
                                                                            p2=p2,
                                                                            tp=tp)
            strings.append(s)
        # bonding matrix
        strings.append("Bonding:\n")
        for (i0,i1),b in numpy.ndenumerate(self.bonds):
            if i0<i1 and b>0:
                s = "{i0:<3} connected to {i1:<3}, bo = {bo:1.2f}\n".format(i0=i0,
                                                                            i1=i1,
                                                                            bo=b)
                strings.append(s)
        return "".join(strings)

    def __str__(self) -> str:
        """Uses repr to print the string.TODO=distinct."""
        return self.__repr__()

    def set_atoms(self,
                  atoms   : ase.Atoms,
                  analyze : bool = False) -> None:
        """Set new Atoms object and reanalyze"""
        logger.debug("Resetting Atoms in SBU {0}".format(self.name))
        self.atoms = atoms
        if analyze:
            logger.debug("\tAnalysis required.")
            self._analyze()
        return None

    def copy(self) -> object:
        """Return a copy of the object"""
        logger.debug("SBU {0}: creating copy.".format(self.name))
        new = SBU(name=str(self.name))
        new.set_atoms(atoms=self.get_atoms(),analyze=False)
        new.symmops = self.symmops.copy()
        new.mmtypes = numpy.copy(self.mmtypes)
        new.bonds   = numpy.copy(self.bonds)
        new.shape   = (self.shape[0],self.shape[1])
        return new

    def get_unique_operations(self) -> dict:
        """Return all unique symmetry operations in the topology."""
        logger.debug("SBU {0}: listing symmetry operations.".format(self.name))
        these_ops = []
        for s,o in self.symmops.items():
            if o is None:
                continue
            elif len(o)>0 and s not in ["I","-I"]:
                sym = ["{0}{1}".format(s,this_o[0]) for this_o in o]
            elif len(o)>0 and o is not None:
                sym = [s,]
            else:
                sym = []
            these_ops += sym
        ops = {o:these_ops.count(o) for o in set(these_ops)}
        return ops

    def is_compatible(self,
                      symmops : list ) -> bool:
        """Return True if symmetry compatible with symmops.
        
        For an SBU to be compatible for alignment with a topology slot
        it has to have at least as many symmetry operators as the slot for
        each type of symmetry. We assume the coordination is checked elsewhere.
        symmops -- (op type, count). e.g: ("C2",5) means 5 C2 in the slot.
        """
        logger.debug("SBU {0}: checking compatibility with slot.".format(self.name))
        compatible = True
        sbuops = self.get_unique_operations()
        for s,o in symmops.items():
            if s not in sbuops.keys():
                compatible = False
            elif o>sbuops[s]:
                 compatible = False
            if not compatible:
                break
        logger.debug("\tCompatibility = {cmp}.".format(cmp=compatible))
        return compatible

    def get_atoms(self) -> ase.Atoms:
        """Return a copy of the topology as ASE Atoms."""
        logger.debug("SBU {0}: returning atoms.".format(self.name))
        return self.atoms.copy()

    def _analyze(self) -> None:
        """Guesses the mmtypes, bonds and pointgroup"""
        logger.debug("SBU {0}: analyze bonding and symmetry.".format(self.name))
        dummies = ase.Atoms([x for x in self.atoms if x.symbol=="X"])
        if len(dummies)>0:
            pg = PointGroup(dummies,0.1)
            self.shape    = (pg.schoenflies,len(dummies))
            self.symmops  = pg.symmops
        bonds,mmtypes = analyze_mm(self.get_atoms())
        self.bonds    = bonds
        self.mmtypes  = mmtypes
        return None

    def transfer_tags(self,
                      fragment : ase.Atoms) -> None:
        """Transfer tags between an aligned fragment and the SBU"""
        logger.debug("\tTagging dummies in SBU {n}.".format(n=self.name))
        # we keep a record of used tags.
        unused = [x.index for x in self.atoms if x.symbol=="X"]
        for atom in fragment:
            ids = [s.index for s in self.atoms if s.index in unused]
            pf = atom.position
            ps = self.atoms.positions[unused]
            d  = numpy.linalg.norm(ps-pf,axis=1)
            si = ids[numpy.argmin(d)]
            self.atoms[si].tag = atom.tag
            unused.remove(si)
        return None

def read_sbu_database(update : bool = False,
                      path   : str  = None):
    """Return a dictionnary of ASE Atoms as SBUs"""
    from autografs.utils.io import read_sbu
    db_file = os.path.join(__data__,"sbu/sbu.pkl")
    user_db = (path is not None)
    no_dflt = (not os.path.isfile(db_file))
    if (user_db or update or no_dflt):
        logger.info("Reloading the building units from scratch")
        sbu = read_sbu(path=path)
        sbu_len = len(sbu)
        logger.info("{0:<5} sbu saved".format(sbu_len))
        with open(db_file,"wb") as pkl:
            pickle.dump(obj=sbu,file=pkl)
        return sbu
    else:
        logger.info("Using saved sbu")
        with open(db_file,"rb") as pkl:
            sbu = pickle.load(file=pkl)
            sbu_len = len(sbu)
            logger.info("{0:<5} sbu loaded".format(sbu_len))
            return sbu


if __name__ == "__main__":

    sbus = read_sbu_database(update=True)

